"""Handles requests and data un-entanglement to a truly retarded web service."""
from __future__ import annotations

import binascii
import logging
from typing import TypeVar, cast
from urllib.parse import urlencode

import aiohttp

from .constants import (
    API_URL,
    CLIENT_TIMEOUT,
    CLIENT_VERSION_NUMBER,
    DEFAULT_COUNTRY,
    NO_ACCESS_RESPONSES,
    NO_CONTENT_RESPONSES,
    PHONE_PREFIXES,
)
from .exceptions import (
    AuthenticationBlockedError,
    AuthenticationError,
    NoAccessError,
    NoContentError,
    NotActivatedError,
)
from .models import (
    Account,
    AccountBase,
    AuthenticationResponse,
    AuthStatus,
    PoliceControlResponse,
    PoliceControlsResponse,
    PoliceControlType,
    PoliceGPSControlsResponse,
    PolitiKontrollerResponse,
    UserMap,
)
from .utils import (
    aes_decrypt,
    aes_encrypt,
    get_query_params,
    map_response_data,
)

ResponseT = TypeVar(
    "ResponseT",
    bound=PolitiKontrollerResponse | dict[str, any],
)

JUNK_CHARS = '\x00\x01\x02\x03\x04\x05\x06\x07\x08\x10\x0f'

_LOGGER = logging.getLogger(__name__)


class Client:
    def __init__(
        self,
        user: Account | None = None,
        session: aiohttp.ClientSession | None = None,
    ):
        self.user = user
        self._web_session = session

    @classmethod
    def initialize(cls, username: str, password: str) -> Client:
        return cls(Account(username=username, password=password))

    @classmethod
    async def login(cls, username: str, password: str) -> Client:
        c = cls()
        await c.authenticate_user(username, password)
        return c

    @property
    def web_session(self):
        if self._web_session:
            return self._web_session
        return aiohttp.ClientSession()

    async def api_request(
        self,
        params: dict,
        headers: dict | None = None,
        cast_to: type[ResponseT] | None = None,
        is_list=False,
    ) -> ResponseT | list[ResponseT] | str:
        method = params.get('p')
        no_auth_methods = ['l', 'r', 'check', 'endre_passord', 'endrings_kode']
        if method not in no_auth_methods and self.user:
            params.update(self.user.get_query_params())
        data = await self.do_external_api_request(params, headers)
        _LOGGER.debug("Got response: %s", data)
        if data in NO_ACCESS_RESPONSES:
            raise NoAccessError
        if data in NO_CONTENT_RESPONSES:
            raise NoContentError

        # Attempt to cast the response data to desired model
        if cast_to is not None:
            model_data = cast_to.from_response_data(data, multiple=is_list)
            if is_list:
                return [cast(cast_to, cast_to.model_validate(d)) for d in model_data]
            return cast(cast_to, cast_to.model_validate(model_data))

        # Return the raw response (str)
        return data

    async def do_external_api_request(
        self,
        params: dict,
        headers: dict | None = None
    ):
        if headers is None:
            headers = {}

        payload = get_query_params(params)
        _LOGGER.debug("Doing API request with params: %s", payload)
        url = f'{API_URL}/app.php?{aes_encrypt(urlencode(payload))}'
        headers = {
            'user-agent': f'PK_{CLIENT_VERSION_NUMBER}',
            **headers,
        }

        async with self.web_session as session:
            async with session.get(
                url,
                headers=headers,
                timeout=aiohttp.ClientTimeout(total=CLIENT_TIMEOUT),
            ) as resp:
                enc_data = await resp.text('utf-8')
                try:
                    data = aes_decrypt(enc_data)
                except binascii.Error:
                    data = enc_data

                return data.strip(JUNK_CHARS).strip()

    def set_user(self, user: Account):
        self.user = user

    async def authenticate_user(self, username: str, password: str):
        auth_user = AccountBase(username=username, password=password)
        params = {
            'p': 'l',
            'lang': auth_user.country.lower(),
            **auth_user.get_query_params(),
        }

        result = await self.api_request(params, cast_to=AuthenticationResponse)
        _LOGGER.debug("Got result: %s", result)

        if result.auth_status == AuthStatus.LOGIN_ERROR:
            raise AuthenticationError
        if result.auth_status == AuthStatus.SPERRET:
            raise AuthenticationBlockedError
        if result.auth_status == AuthStatus.NOT_ACTIVATED:
            raise NotActivatedError

        account_dict = {
            **auth_user.model_dump(),
            **result.model_dump(),
            **{"username": f"{auth_user.username}"},  # noqa: PIE800
        }

        account = Account(**{str(k): str(v) for k, v in account_dict.items()})
        self.set_user(account)
        return account

    async def get_settings(self):
        params = {
            'p': 'instillinger',
        }
        return await self.api_request(params)

    async def get_control(self, cid: int) -> PoliceControlResponse:
        params = {
            'p': 'hki',
            'kontroll_id': cid,
        }
        return await self.api_request(params, cast_to=PoliceControlResponse)

    async def get_controls(
        self,
        lat: float,
        lng: float,
        **kwargs,
    ) -> list[PoliceControlsResponse]:
        params = {
            'p': 'hk',
            'lat': lat,
            'lon': lng,
            **kwargs,
        }

        try:
            return await self.api_request(
                params,
                cast_to=PoliceControlsResponse,
                is_list=True,
            )
        except NoContentError:
            return []

    async def get_controls_in_radius(
        self,
        lat: float,
        lng: float,
        radius: int,
        speed: int = 100,
        **kwargs,
    ) -> list[PoliceGPSControlsResponse]:
        params = {
            'p': 'gps_kontroller',
            'vr': radius,
            'speed': speed,
            'lat': lat,
            'lon': lng,
            **kwargs,
        }

        try:
            return await self.api_request(
                params,
                cast_to=PoliceGPSControlsResponse,
                is_list=True,
            )
        except NoContentError:
            return []

    async def get_controls_from_lists(
        self,
        controls: list[PoliceGPSControlsResponse | PoliceControlsResponse],
    ) -> list[PoliceControlResponse]:
        return [await self.get_control(i.id) for i in controls]

    async def get_control_types(self) -> list[PoliceControlType]:
        params = {
            'p': 'kontrolltyper',
        }
        return await self.api_request(params, cast_to=PoliceControlType, is_list=True)

    async def get_maps(self) -> list[UserMap]:
        params = {
            'p': 'hent_mine_kart',
        }
        return await self.api_request(params, cast_to=UserMap, is_list=True)

    async def exchange_points(self):
        params = {
            'p': 'veksle',
        }
        result = await self.api_request(params)
        return map_response_data(result, [
            'status',
            'message',
        ])

    async def account_register(
        self,
        phone_number: int,
        password: str,
        name: str,
        country: str | None = None,
    ):
        if country is None:
            country = DEFAULT_COUNTRY
        country_code = PHONE_PREFIXES.get(country, 00)

        params = {
            'p': 'r',
            'telefon': phone_number,
            'passord': password,
            'cc': country_code,
            'navn': name,
            'lang': country,
        }
        result = await self.api_request(params)
        return map_response_data(result, [
            'status',
            'message',
        ])

    async def account_auth(self, auth_code: str, uid: int):
        params = {
            'p': 'auth_app',
            'auth_kode': auth_code,
            'uid': uid,
        }
        result = await self.api_request(params)
        return map_response_data(result, [
            'status',
            'message',
        ])

    async def account_auth_sms(self):
        params = {
            'p': 'auth_sms',
        }
        result = await self.api_request(params)
        return map_response_data(result, [
            'status',
            'message',
        ])

    async def account_send_sms(self):
        params = {
            'p': 'send_sms',
        }
        result = await self.api_request(params)
        return map_response_data(result, [
            'status',
            'message',
        ])



"""Politikontroller models."""
from __future__ import annotations

from abc import ABC
from datetime import datetime
from enum import Enum, auto
import logging
from typing import ClassVar, Literal

from pydantic import BaseModel, ConfigDict, field_validator

from .constants import DEFAULT_COUNTRY, PHONE_NUMBER_LENGTH, PHONE_PREFIXES
from .utils import map_response_data, parse_time_format

_LOGGER = logging.getLogger(__name__)


# noinspection PyUnresolvedReferences
class StrEnum(str, Enum):
    """A string enumeration of type `(str, Enum)`.
    All members are compared via `upper()`. Defaults to UNKNOWN.
    """

    def __eq__(self, other: str) -> bool:
        other = other.upper()
        return super().__eq__(other)

    @classmethod
    def _missing_(cls, value) -> str:
        has_unknown = False
        for member in cls:
            if member.name.upper() == "unknown":
                has_unknown = True
            if member.name.upper() == value.upper():
                return member
        if has_unknown:
            _LOGGER.warning("'%s' is not a valid '%s'", value, cls.__name__)
            return cls.UNKNOWN
        raise ValueError(f"'{value}' is not a valid {cls.__name__}")


class AuthStatus(StrEnum):
    APP_ERR = 'APP_ERR'
    LOGIN_OK = 'LOGIN_OK'
    LOGIN_ERROR = 'LOGIN_ERROR'
    SPERRET = 'SPERRET'
    NOT_ACTIVATED = 'NOT_ACTIVATED'
    SKIP_AUTHENTICATION = 'SKIP_AUTHENTICATION'


class ExchangeStatus(StrEnum):
    EXCHANGE_OK = auto()


class PolitiKontrollerRequest(BaseModel):
    pass


class PolitiKontrollerResponse(BaseModel, ABC):
    model_config = ConfigDict(use_enum_values=True)

    attr_map: ClassVar[list[str]]

    @classmethod
    def from_response_data(
        cls,
        data: str,
        multiple=False
    ) -> list[dict[str, str]] | dict[str, str]:
        """Convert a cvs-like string into dictionaries."""
        return map_response_data(data, cls.attr_map, multiple)


class AuthenticationResponse(PolitiKontrollerResponse):
    auth_status: AuthStatus
    premium_key: str = "NO"
    user_level: int
    phone_prefix: int
    status: str | None = None
    uid: int
    nickname: str | None = None
    saphne: Literal["SAPHE", "NO_SAPHE"] | None = None
    show_regnr: Literal["REGNR", "NO_REGNR"] | None = None
    premium_price: int | None = None
    enable_points: Literal["YES", "NO"] | None = None
    enable_calls: Literal["YES", "NO"] | None = None
    needs_gps: bool | None = None
    gps_radius: int | None = None
    push_notification: bool | None = None
    sms_notification: bool | None = None
    points: int | None = None
    exchange_code: bool | None = None

    attr_map = [
        'auth_status',
        # 0  LoginStatus:  APP_ERR|LOGIN_OK|NOT_ACTIVATED|SPERRET|LOGIN_ERROR
        'premium_key',  # 1  PremiumKey: str | Literal["NO"]
        'user_level',  # 2  user_level: int
        'phone_prefix',  # 3  RetningsKode: int
        'status',  # 4  status: str
        'uid',  # 5  brukerId: int
        None,  # 6
        'nickname',  # 7  kallenavn: str
        'saphne',  # 8  saphne: Literal["SAPHE"] | None
        'show_regnr',  # 9  vis_regnr: Literal["REGNR"] | None
        'premium_price',  # 10 premium_price: int
        'enable_points',  # 11 enable_points: str
        'enable_calls',  # 12 enable_calls: str
        None,  # 13 needs_gps: bool
        'gps_radius',  # 14 gps_radius: int
        'push_notification',  # 15 push_varsling: bool
        'sms_notification',  # 16 sms_varsling: bool
        'points',  # 17 DinePoeng: int
        'exchange_code',  # 18 LosInnKode: bool
    ]


class PoliceControlTypeEnum(str, Enum):
    SPEED_TRAP = "Fartskontroll"
    BEHAVIOUR = "Belte/mobil"
    TECHNICAL = "Teknisk"
    TRAFFIC_INFO = "Trafikk info"
    TRAFFIC_MESSAGE = "Trafikkmelding"
    OBSERVATION = "Observasjon"
    CUSTOMS = "Toll/grense"
    WEIGHT = "Vektkontroll"
    UNKNOWN = "Ukjent"
    CIVIL_POLICE = "Sivilpoliti"
    MC_CONTROL = "Mopedkontroll"
    BOAT_PATROL = "Politibåten"


class BaseResponseModel(BaseModel):
    pass


class AuthenticationRequest(PolitiKontrollerRequest):
    username: str
    password: str


# noinspection PyNestedDecorators
class AccountBase(BaseModel):
    username: str
    password: str | None = None
    country: str = DEFAULT_COUNTRY

    @property
    def phone_number(self):
        return int(self.username[2:]) \
            if len(self.username) > PHONE_NUMBER_LENGTH \
            else int(self.username)

    @property
    def phone_prefix(self):
        return int(self.username[:2]) \
            if len(self.username) > PHONE_NUMBER_LENGTH \
            else PHONE_PREFIXES.get(self.country.lower())

    @field_validator('username', mode='before')
    @classmethod
    def validate_username(cls, v: str):
        return str(v).replace(' ', '')

    def get_query_params(self):
        """Get query params."""
        return {
            'retning': self.phone_prefix,
            'telefon': self.phone_number,
            'passord': self.password,
        }


class Account(AccountBase):
    uid: int | None
    auth_status: AuthStatus | None
    status: str | None


# noinspection PyNestedDecorators
class PoliceControlType(PolitiKontrollerResponse):
    id: int
    name: PoliceControlTypeEnum
    slug: str

    attr_map = [
        'slug',
        'name',
        'id',
        None,
    ]

    @field_validator('slug', mode='before')
    @classmethod
    def validate_slug(cls, v: str):
        # Remove ".png"
        return v[:-4]


class PoliceControlPoint:
    type: str = "Point"
    lat: float
    lng: float

    def __init__(self, lat, lng):
        self.lat = lat
        self.lng = lng

    @property
    def coordinates(self):
        return self.lng, self.lat

    @property
    def __geo_interface__(self):
        return {
            'type': self.type,
            'coordinates': self.coordinates,
        }


# noinspection PyNestedDecorators
class PoliceControlResponse(PolitiKontrollerResponse):
    id: int
    county: str
    municipality: str
    type: PoliceControlTypeEnum
    timestamp: datetime | None
    description: str
    lat: float
    lng: float
    speed_limit: int | None = None
    last_seen: datetime | None
    confirmed: int = 0

    attr_map = [
        'id',  # 0  id: int  14241
        'county',  # 1  country: str       Trøndelag
        'municipality',  # 2  municipality: str   Malvik
        'type',  # 3  control_type: str   Fartskontroll
        'timestamp',  # 4     29.05 - 20:47
        'description',  # 5     Kontroll Olderdalen
        'lat',  # 6     63.4258007013951
        'lng',  # 7     10.6856604194473
        None,  # 8     |
        None,  # 9     |
        None,  # 10    malvik.png
        None,  # 11    trondelag.png
        'speed_limit',  # 12 speed_limit: int   90
        None,  # 13 enabled   1
        'last_seen',  # 14 last_seen: time   20:47
        'confirmed',  # 15 confirmed: str    0  (if not 0: confirmed=red)
        None,  # 16 confirmed   2   (0=green,  1=orange, 2=red)
        None,  # 17 control_type: int   1
    ]

    @field_validator('timestamp', mode='before')
    @classmethod
    def validate_timestamp(cls, v: int):
        return cls._timestamp_validate(v)

    @field_validator('last_seen', mode='before')
    @classmethod
    def validate_last_seen(cls, v: int):
        return cls._timestamp_validate(v)

    @property
    def description_truncated(self):
        return (
            self.description[:25] + '..'
        ) if len(self.description) > 27 else self.description  # noqa: PLR2004

    @property
    def title(self):
        return f"{self.type}: {self.description_truncated}"

    @classmethod
    def _timestamp_validate(cls, v: str | int) -> int | str | None:
        if len(v) == 0 or (v.isnumeric() and int(v) == 0):
            return None
        return parse_time_format(v)

    @property
    def _geometry(self):
        return PoliceControlPoint(self.lat, self.lng)

    @property
    def __geo_interface__(self):
        return {
            "type": "Feature",
            "geometry": self._geometry.__geo_interface__,
            "properties": {
                "title": self.title,
                "description": self.description,
                "type": self.type,
            },
        }


# noinspection PyNestedDecorators
class PoliceGPSControlsResponse(PolitiKontrollerResponse):
    id: int
    county: str
    municipality: str
    type: PoliceControlTypeEnum
    timestamp: datetime | None
    description: str
    lat: float
    lng: float

    attr_map = [
        'id',
        'county',
        'municipality',
        'type',
        'timestamp',
        'description',
        'lat',
        'lng',
        None,
        None,
        None,
        None,
    ]

    @field_validator('timestamp', mode='before')
    @classmethod
    def validate_timestamp(cls, v: int):
        return cls._timestamp_validate(v)

    @classmethod
    def _timestamp_validate(cls, v: str | int) -> int | str | None:
        if len(v) == 0 or (v.isnumeric() and int(v) == 0):
            return None
        return parse_time_format(v)


# noinspection PyNestedDecorators
class PoliceControlsResponse(PolitiKontrollerResponse):
    id: int
    county: str
    municipality: str
    type: PoliceControlTypeEnum
    timestamp: datetime | None
    description: str
    lat: float
    lng: float
    last_seen: datetime | None

    attr_map = [
        'id',
        'county',
        'municipality',
        'type',
        None,
        'description',
        'lat',
        'lng',
        None,
        None,
        None,
        None,
        'timestamp',
        None,
        None,
        'last_seen',
    ]

    @field_validator('timestamp', mode='before')
    @classmethod
    def validate_timestamp(cls, v: int):
        return cls._timestamp_validate(v)

    @field_validator('last_seen', mode='before')
    @classmethod
    def validate_last_seen(cls, v: int):
        return cls._timestamp_validate(v)

    @classmethod
    def _timestamp_validate(cls, v: str | int) -> int | str | None:
        if len(v) == 0 or (v.isnumeric() and int(v) == 0):
            return None
        return parse_time_format(v)


class UserMap(PolitiKontrollerResponse):
    id: int
    title: str
    country: str

    attr_map = [
        'id',
        None,
        'title',
        'country',
    ]


class ExchangePointsResponse(PolitiKontrollerResponse):
    status: ExchangeStatus
    message: str



class PolitikontrollerError(Exception):
    pass


class NoAccessError(PolitikontrollerError):
    pass


class NoContentError(PolitikontrollerError):
    pass


class AuthenticationError(PolitikontrollerError):
    pass


class AuthenticationBlockedError(AuthenticationError):
    pass


class NotActivatedError(AuthenticationError):
    pass



"""CLI tool."""

import logging

import anyio
import asyncclick as click
from asyncclick.core import Context
from tabulate import tabulate

from politikontroller_py import Client
from politikontroller_py.exceptions import AuthenticationError

TABULATE_DEFAULTS = {
    'tablefmt': 'rounded_grid',
}


@click.group()
@click.option('--username', '-u', envvar='POLITIKONTROLLER_USERNAME', type=str,
              required=True, help='Username (i.e. phone number)')
@click.password_option('--password',
                       '-p',
                       envvar='POLITIKONTROLLER_PASSWORD',
                       type=str,
                       required=True,
                       confirmation_prompt=False,
                       help='Password')
@click.option('--debug', is_flag=True, help='Set logging level to DEBUG')
@click.pass_context
async def cli(ctx: Context, username: str, password: str, debug: bool):
    """Username and password can be defined using env vars:

    POLITIKONTROLLER_USERNAME
    POLITIKONTROLLER_PASSWORD
    """
    configure_logging(debug)

    ctx.obj = client = Client()

    try:
        await client.authenticate_user(username, password)
    except AuthenticationError as err:
        raise click.BadParameter(str(err), param_hint="--username, --password") from err


@cli.command('get-controls', short_help='get a list of all active controls.')
@click.option('--lat', required=True, help='Your position (latitude)')
@click.option('--lng', required=True, help='Your position (longitude)')
@click.pass_obj
async def get_controls(obj, lat: float, lng: float):
    controls = await obj.get_controls(lat, lng)
    click.echo(tabulate(controls, **TABULATE_DEFAULTS))


@cli.command('get-controls-radius',
             short_help='get all active controls inside a radius.')
@click.option('--lat', type=float, required=True, help='Radius center (latitude)')
@click.option('--lng', type=float, required=True, help='Radius center (longitude)')
@click.option('--radius', type=int, required=True, metavar='km',
              help='Radius size in kilometers')
@click.option('--speed', type=int, required=False, metavar='km/h',
              help='Speed, unknown what this does')
@click.pass_obj
async def get_controls_in_radius(obj: Client, lat: float, lng: float, radius: int, speed: int):
    _controls = await obj.get_controls_in_radius(lat, lng, radius, speed)
    controls = await obj.get_controls_from_lists(_controls)

    click.echo(tabulate([d.model_dump() for d in controls], **TABULATE_DEFAULTS))


@cli.command('get-control', short_help='get details on a control.')
@click.argument('control_id', type=int, required=True)
@click.pass_obj
async def get_control(obj: Client, control_id: int):
    control = await obj.get_control(control_id)
    click.echo(tabulate(control, **TABULATE_DEFAULTS))


@cli.command('get-maps', short_help='get own maps.')
@click.pass_obj
async def get_maps(obj: Client):
    maps = await obj.get_maps()
    print(maps)


@cli.command('exchange-points', short_help='exchange points (?)')
@click.pass_obj
async def exchange_points(obj: Client):
    res = await obj.exchange_points()
    print(res)


@cli.command('account-send-sms', short_help='send activation sms')
@click.pass_obj
async def account_send_sms(obj: Client):
    res = await obj.account_send_sms()
    print(res)


@cli.command('account-auth-sms', short_help='activate account by sms')
@click.pass_obj
async def account_auth_sms(obj: Client):
    res = await obj.account_auth_sms()
    print(res)


@cli.command('account-auth', short_help='activate account')
@click.argument('code', type=str, required=True)
@click.argument('uid', type=int, required=True)
@click.pass_obj
async def account_auth(obj: Client, code: str, uid: int):
    res = await obj.account_auth(auth_code=code, uid=uid)
    print(res)


@cli.command('account-register', short_help='register new account')
@click.argument('phone', type=int, required=True)
@click.argument('password', type=str, required=True)
@click.argument('name', type=str, required=True)
@click.argument('country', type=str, required=True)
@click.pass_obj
async def account_register(
    obj: Client,
    phone: int,
    password: str,
    name: str,
    country: str,
):
    res = await obj.account_register(
        phone_number=phone,
        password=password,
        name=name,
        country=country,
    )
    print(res)


def configure_logging(debug: bool = False):
    level = logging.DEBUG if debug else logging.INFO
    logging.basicConfig(level=level)


async def main():
    await cli.main()


if __name__ == '__main__':
    # noinspection PyTypeChecker
    anyio.run(main)
